#include "sd/malloc.h"
#include "sd/test.h"

#include <io.h>
#include <process.h> /* for system() */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define sd_malloc	malloc
#define sd_calloc	calloc
#define sd_realloc	realloc
#define sd_strdup	strdup

#define DIFF_CMD  "/usr/bin/diff -q"

#define MAX_NFUNC 100

typedef long usec_t;

struct __sd_test
{
    const char*         name;
    char                in_filename[128];
    char                ref_filename[128];
    char                out_filename[128];
    FILE*               in;
    FILE*               out;
    FILE*               err;
    int                 verbose;
    int                 timed;
    sd_test_func_t**    funcs;
    int                 size;
    int			argc;
    char**		argv;
};

/**
*******************************************************************************

@brief  return current time.[usec]
*/
static usec_t now()
{
       time_t   ret;

    time( &ret );


    return ret;

}

#define R_OK 00

/******************************************************************************/
static int test_compare(sd_test_t* pme, int a_argc, char* a_argv[])
{
    char cmd[1024];

    if (access(pme->ref_filename, R_OK) || access(pme->out_filename, R_OK))
        return 1;

    _snprintf(cmd, sizeof(cmd), "%s %s %s 1>/dev/null 2>&1", DIFF_CMD,
             pme->ref_filename, pme->out_filename);

    return ! system(cmd);
}



/******************************************************************************/
extern sd_test_t* sd_test_new(int a_argc, char* a_argv[])
{
    sd_test_t* pme;
    int c;

    sd_test_func_t dummy;
    
    c = sizeof( (size_t)dummy );
    pme        = sd_calloc(1, sizeof(sd_test_t));
    pme->funcs = sd_calloc(MAX_NFUNC, c);
    pme->name  = strstr(a_argv[0], "lt-") ? a_argv[0] + 3 : a_argv[0];

    _snprintf(pme->ref_filename, sizeof(pme->ref_filename), "%s.ref",
	     pme->name);
    _snprintf(pme->in_filename,  sizeof(pme->in_filename),  "%s.in",
	     pme->name);
    _snprintf(pme->out_filename, sizeof(pme->out_filename), "%s.out",
	     pme->name);
    
    pme->in    = fopen(pme->in_filename,  "r");
    pme->out   = fopen(pme->out_filename, "w");
    pme->err   = 0;
    pme->verbose= 0;
    pme->size  = 0;

    /*
    while ((c = getopt(a_argc, a_argv, "vt")) != EOF) {
        switch(c) {
        case 'v': pme->verbose = 1; break;
        case 't': pme->timed   = 1; break;
        default:                     break; 
        }
    }

    pme->argc = a_argc - (optind - 1);
    pme->argv = a_argv + (optind - 1);
    */
    return pme;
}


/******************************************************************************/
extern void sd_test_delete(sd_test_t* pme)
{
    if (!pme)
        return;

    if (pme->in) fclose(pme->in);
    if (pme->out) fclose(pme->out);
    free(pme->funcs);
    free(pme);
}
        
/******************************************************************************/
extern const char* sd_test_get_name(const sd_test_t* pme)
{
    return pme ? pme->name : NULL;
}

/******************************************************************************/
extern int sd_test_get_verbose(const sd_test_t* pme)
{
    return pme ? pme->verbose: 0;
}

/******************************************************************************/
extern int sd_test_set_verbose(sd_test_t* pme, int a_verbose)
{
    if (!pme)
        return 0;

    return pme->verbose = a_verbose;
}

/******************************************************************************/
extern int sd_test_get_timed(const sd_test_t* pme)
{
    return pme ? pme->timed: 0;
}

/******************************************************************************/
extern int sd_test_set_timed(sd_test_t* pme, int a_timed)
{
    if (!pme)
        return 0;

    return pme->timed = a_timed;
}
/******************************************************************************/
extern FILE* sd_test_in(sd_test_t* pme)
{
    if (!pme)
        return NULL;

    return pme->in ? pme->in : stdin;
}

/******************************************************************************/
extern FILE* sd_test_out(sd_test_t* pme)
{
    if (!pme)
        return NULL;

    if (pme->verbose)
        return stdout;

    return pme->out ? pme->out : stdout;
}

/******************************************************************************/
extern FILE* sd_test_err(sd_test_t* pme)
{
    if (!pme)
        return NULL;

    return pme->err ? pme->err : stderr;
}

/******************************************************************************/
extern int sd_test_run(sd_test_t* pme, int argc, char* argv[])
{
    int i, passed = 0;

    if (!pme)
        return -1;

    sd_test_add(pme, test_compare);
    
    fprintf(sd_test_err(pme), "%s: ", pme->name);
    
    for (i = 0; i < pme->size; i++) {
        int t;
        usec_t elapsed;

        fprintf(sd_test_out(pme), "=> test #%d :\n", i);

        elapsed = now();
        t = (*pme->funcs[i])(pme, pme->argc, pme->argv);
        elapsed = now() - elapsed;

        fprintf(sd_test_out(pme), "=> test #%d : %s\n", i, t ? " passed" : " failed");
        fflush(sd_test_out(pme));

        passed += (t != 0);
        
        fprintf(sd_test_err(pme), "%c", t ? '+' : '-');
        if (pme->timed)
            fprintf(sd_test_err(pme), "%lld ", elapsed);
    }
    
    fprintf(sd_test_err(pme), " %d/%d %s.\n", passed, pme->size,
            passed == pme->size ? " passed" : " failed");
    
    return passed == pme->size;
}

/******************************************************************************/
extern int sd_test_add(sd_test_t* pme, sd_test_func_t a_func)
{
    if (!pme)
        return -1;

    if (pme->size == MAX_NFUNC)
        return pme->size;

    pme->funcs[pme->size] = a_func;

    return pme->size++;
}
