#ifndef _QW_mutex_h
#define _QW_mutex_h

/* Functions to provide thread synchronization primitives

	These are independent of the other QW routines.
*/

#include "QW_main.h"
#include "QW_types.h"

#include "begin_code.h"
/* Set up for C function definitions, even when using C++ */
#ifdef __cplusplus
extern "C" {
#endif

/* Synchronization functions which can time out return this value
   if they time out.
*/
#define QW_MUTEX_TIMEDOUT	1

/* This is the timeout value which corresponds to never time out */
#define QW_MUTEX_MAXWAIT	(~(Uint32)0)


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Mutex functions                                               */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* The QW mutex structure, defined in QW_mutex.c */
struct QW_mutex;
typedef struct QW_mutex QW_mutex;

/* Create a mutex, initialized unlocked */
extern DECLSPEC QW_mutex * QWCALL QW_CreateMutex(void);

/* Lock the mutex  (Returns 0, or -1 on error) */
#define QW_LockMutex(m)	QW_mutexP(m)
extern DECLSPEC int QWCALL QW_mutexP(QW_mutex *mutex);

/* Unlock the mutex  (Returns 0, or -1 on error)
   It is an error to unlock a mutex that has not been locked by
   the current thread, and doing so results in undefined behavior.
 */
#define QW_UnlockMutex(m)	QW_mutexV(m)
extern DECLSPEC int QWCALL QW_mutexV(QW_mutex *mutex);

/* Destroy a mutex */
extern DECLSPEC void QWCALL QW_DestroyMutex(QW_mutex *mutex);


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Semaphore functions                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* The QW semaphore structure, defined in QW_sem.c */
struct QW_semaphore;
typedef struct QW_semaphore QW_sem;

/* Create a semaphore, initialized with value, returns NULL on failure. */
extern DECLSPEC QW_sem * QWCALL QW_CreateSemaphore(Uint32 initial_value);

/* Destroy a semaphore */
extern DECLSPEC void QWCALL QW_DestroySemaphore(QW_sem *sem);

/* This function suspends the calling thread until the semaphore pointed 
 * to by sem has a positive count. It then atomically decreases the semaphore
 * count.
 */
extern DECLSPEC int QWCALL QW_SemWait(QW_sem *sem);

/* Non-blocking variant of QW_SemWait(), returns 0 if the wait succeeds,
   QW_MUTEX_TIMEDOUT if the wait would block, and -1 on error.
*/
extern DECLSPEC int QWCALL QW_SemTryWait(QW_sem *sem);

/* Variant of QW_SemWait() with a timeout in milliseconds, returns 0 if
   the wait succeeds, QW_MUTEX_TIMEDOUT if the wait does not succeed in
   the allotted time, and -1 on error.
   On some platforms this function is implemented by looping with a delay
   of 1 ms, and so should be avoided if possible.
*/
extern DECLSPEC int QWCALL QW_SemWaitTimeout(QW_sem *sem, Uint32 ms);

/* Atomically increases the semaphore's count (not blocking), returns 0,
   or -1 on error.
 */
extern DECLSPEC int QWCALL QW_SemPost(QW_sem *sem);

/* Returns the current count of the semaphore */
extern DECLSPEC Uint32 QWCALL QW_SemValue(QW_sem *sem);


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Condition variable functions                                  */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* The QW condition variable structure, defined in QW_cond.c */
struct QW_cond;
typedef struct QW_cond QW_cond;

/* Create a condition variable */
extern DECLSPEC QW_cond * QWCALL QW_CreateCond(void);

/* Destroy a condition variable */
extern DECLSPEC void QWCALL QW_DestroyCond(QW_cond *cond);

/* Restart one of the threads that are waiting on the condition variable,
   returns 0 or -1 on error.
 */
extern DECLSPEC int QWCALL QW_CondSignal(QW_cond *cond);

/* Restart all threads that are waiting on the condition variable,
   returns 0 or -1 on error.
 */
extern DECLSPEC int QWCALL QW_CondBroadcast(QW_cond *cond);

/* Wait on the condition variable, unlocking the provided mutex.
   The mutex must be locked before entering this function!
   Returns 0 when it is signaled, or -1 on error.
 */
extern DECLSPEC int QWCALL QW_CondWait(QW_cond *cond, QW_mutex *mut);

/* Waits for at most 'ms' milliseconds, and returns 0 if the condition
   variable is signaled, QW_MUTEX_TIMEDOUT if the condition is not
   signaled in the allotted time, and -1 on error.
   On some platforms this function is implemented by looping with a delay
   of 1 ms, and so should be avoided if possible.
*/
extern DECLSPEC int QWCALL QW_CondWaitTimeout(QW_cond *cond, QW_mutex *mutex, Uint32 ms);

/* Ends C function definitions when using C++ */
#ifdef __cplusplus
}
#endif
#include "close_code.h"

#endif /* _QW_mutex_h */
