/* System independent thread management routines for QW */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "QW_error.h"
#include "QW_mutex.h"
#include "QW_thread.h"
#include "QW_thread_c.h"
#include "QW_systhread.h"

#define ARRAY_CHUNKSIZE	32
/* The array of threads currently active in the application
   (except the main thread)
   The manipulation of an array here is safer than using a linked list.
*/
static int QW_maxthreads = 0;
static int QW_numthreads = 0;
static QW_Thread **QW_Threads = NULL;
static QW_mutex *thread_lock = NULL;
int _creating_thread_lock = 0;

int QW_ThreadsInit(void)
{
	int retval;

	retval = 0;
	/* Set the thread lock creation flag so that we can reuse an
	   existing lock on the system - since this mutex never gets
	   destroyed (see QW_ThreadsQuit()), we want to reuse it.
	*/
	_creating_thread_lock = 1;
	thread_lock = QW_CreateMutex();
	_creating_thread_lock = 0;
	if ( thread_lock == NULL ) {
		retval = -1;
	}
	return(retval);
}

/* This should never be called...
   If this is called by QW_Quit(), we don't know whether or not we should
   clean up threads here.  If any threads are still running after this call,
   they will no longer have access to any per-thread data.
 */
void QW_ThreadsQuit()
{
	QW_mutex *mutex;

	mutex = thread_lock;
	thread_lock = NULL;
	if ( mutex != NULL ) {
		QW_DestroyMutex(mutex);
	}
}

/* Routines for manipulating the thread list */
static void QW_AddThread(QW_Thread *thread)
{
	QW_Thread **threads;

	/* WARNING:
	   If the very first threads are created simultaneously, then
	   there could be a race condition causing memory corruption.
	   In practice, this isn't a problem because by definition there
	   is only one thread running the first time this is called.
	*/
	if ( thread_lock == NULL ) {
		if ( QW_ThreadsInit() < 0 ) {
			return;
		}
	}
	QW_mutexP(thread_lock);

	/* Expand the list of threads, if necessary */
#ifdef DEBUG_THREADS
	printf("Adding thread (%d already - %d max)\n",
			QW_numthreads, QW_maxthreads);
#endif
	if ( QW_numthreads == QW_maxthreads ) {
		threads=(QW_Thread **)malloc((QW_maxthreads+ARRAY_CHUNKSIZE)*
		                              (sizeof *threads));
		if ( threads == NULL ) {
			QW_OutOfMemory();
			goto done;
		}
		memcpy(threads, QW_Threads, QW_numthreads*(sizeof *threads));
		QW_maxthreads += ARRAY_CHUNKSIZE;
		if ( QW_Threads ) {
			free(QW_Threads);
		}
		QW_Threads = threads;
	}
	QW_Threads[QW_numthreads++] = thread;
done:
	QW_mutexV(thread_lock);
}

static void QW_DelThread(QW_Thread *thread)
{
	int i;

	if ( thread_lock ) {
		QW_mutexP(thread_lock);
		for ( i=0; i<QW_numthreads; ++i ) {
			if ( thread == QW_Threads[i] ) {
				break;
			}
		}
		if ( i < QW_numthreads ) {
			if ( --QW_numthreads > 0 ) {
				while ( i < QW_numthreads ) {
					QW_Threads[i] = QW_Threads[i+1];
					++i;
				}
			} else {
				QW_maxthreads = 0;
				free(QW_Threads);
				QW_Threads = NULL;
			}
#ifdef DEBUG_THREADS
			printf("Deleting thread (%d left - %d max)\n",
					QW_numthreads, QW_maxthreads);
#endif
		}
		QW_mutexV(thread_lock);
	}
}

/* The default (non-thread-safe) global error variable */
static QW_error QW_global_error;

/* Routine to get the thread-specific error variable */
QW_error *QW_GetErrBuf(void)
{
	QW_error *errbuf;

	errbuf = &QW_global_error;
	if ( QW_Threads ) {
		int i;
		Uint32 this_thread;

		this_thread = QW_ThreadID();
		QW_mutexP(thread_lock);
		for ( i=0; i<QW_numthreads; ++i ) {
			if ( this_thread == QW_Threads[i]->threadid ) {
				errbuf = &QW_Threads[i]->errbuf;
				break;
			}
		}
		QW_mutexV(thread_lock);
	}
	return(errbuf);
}


/* Arguments and callback to setup and run the user thread function */
typedef struct {
	int (*func)(void *);
	void *data;
	QW_Thread *info;
	QW_sem *wait;
} thread_args;

void QW_RunThread(void *data)
{
	thread_args *args;
	int (*userfunc)(void *);
	void *userdata;
	int *statusloc;

	/* Perform any system-dependent setup
	   - this function cannot fail, and cannot use QW_SetError()
	 */
	QW_SYS_SetupThread();

	/* Get the thread id */
	args = (thread_args *)data;
	args->info->threadid = QW_ThreadID();

	/* Figure out what function to run */
	userfunc = args->func;
	userdata = args->data;
	statusloc = &args->info->status;

	/* Wake up the parent thread */
	QW_SemPost(args->wait);

	/* Run the function */
	*statusloc = userfunc(userdata);
}

QW_Thread *QW_CreateThread(int (*fn)(void *), void *data)
{
	QW_Thread *thread;
	thread_args *args;
	int ret;

	/* Allocate memory for the thread info structure */
	thread = (QW_Thread *)malloc(sizeof(*thread));
	if ( thread == NULL ) {
		QW_OutOfMemory();
		return(NULL);
	}
	memset(thread, 0, (sizeof *thread));
	thread->status = -1;

	/* Set up the arguments for the thread */
	args = (thread_args *)malloc(sizeof(*args));
	if ( args == NULL ) {
		QW_OutOfMemory();
		free(thread);
		return(NULL);
	}
	args->func = fn;
	args->data = data;
	args->info = thread;
	args->wait = QW_CreateSemaphore(0);
	if ( args->wait == NULL ) {
		free(thread);
		free(args);
		return(NULL);
	}

	/* Add the thread to the list of available threads */
	QW_AddThread(thread);

	/* Create the thread and go! */
	ret = QW_SYS_CreateThread(thread, args);
	if ( ret >= 0 ) {
		/* Wait for the thread function to use arguments */
		QW_SemWait(args->wait);
	} else {
		/* Oops, failed.  Gotta free everything */
		QW_DelThread(thread);
		free(thread);
		thread = NULL;
	}
	QW_DestroySemaphore(args->wait);
	free(args);

	/* Everything is running now */
	return(thread);
}

void QW_WaitThread(QW_Thread *thread, int *status)
{
	if ( thread ) {
		QW_SYS_WaitThread(thread);
		if ( status ) {
			*status = thread->status;
		}
		QW_DelThread(thread);
		free(thread);
	}
}

Uint32 QW_GetThreadID(QW_Thread *thread)
{
	Uint32 id;

	if ( thread ) {
		id = thread->threadid;
	} else {
		id = QW_ThreadID();
	}
	return(id);
}

void QW_KillThread(QW_Thread *thread)
{
	if ( thread ) {
		QW_SYS_KillThread(thread);
		QW_WaitThread(thread, NULL);
	}
}

